import ee
import numpy as np
import pandas as pd
from datetime import datetime

def filter_collection(collection: str, date_range: list, bounds: ee.Geometry) -> str:
    """
    Get a list (collection) of ee images filtered by date range and ee geometry (point or so).
    Input:
    -------
    collection: str, ee collection name
    date_range: list, [start: str, end: str]
    bounds: ee.Geometry, e.g.
        ee.Geometry.Polygon(
        [[
        [minlon, maxlat],
        [maxlon, maxlat],
        [maxlon, minlat],
        [minlon, minlat]
        ]]
        ) 
    Return:
    -------
    A list of ee images (collection)
    """
    collection = ee.ImageCollection(collection)
    collection = collection.filterDate(*date_range)
    collection = collection.filterBounds(bounds)
    return collection.toList(collection.size())

def get_collection_size(collection: ee.ImageCollection) -> int:
    """
    Get the length of collection list
    Input:
    -------
    collecton: ee.ImageCollection
    Return:
    -------
    int, size of collection
    """
    return collection.size().getInfo()

def get_image(collection: ee.ImageCollection, count) -> ee.Image:
    """
    Get the ith image from collection
    Input:
    -------
    collection: ee.ImageCollection, a list of images
    count: index of wanted image in collection
    Return:
    -------
    ee.Image
    """
    image = collection.get(count)
    image = ee.Image(image)
    return image

def get_image_bandnames(image: ee.Image) -> list:
    """
    Get the band names of image
    Input:
    -------
    image: ee.Image
    Return:
    -------
    list, a list of band names
    """
    return image.bandNames().getInfo()

def get_proj(image: ee.Image, band: str, pyfmt: bool):
    """
    Get the projection information of an image
    Input:
    -------
    image: ee.Image
    band: str, band name
    pyfmt: bool, python or ee format of output projection
    Return:
    -------
    Projection of image
    """
    proj = image.select(band).projection()
    if pyfmt:
        proj = proj.getInfo()
    return proj
 
def proj_epsg(epsg: int or str = 4326) -> ee.Projection:
    proj = ee.Projection(f'EPSG:{epsg}')
    return proj
 
def get_scale(image: ee.Image) -> float:
    scale = image.projection().nominalScale().getInfo()
    return scale
 
def get_date(image):
    date = ee.Date(image.get('system:time_start'))
    return date.format('Y-M-d HH:mm:ss.SSSS').getInfo()

def gee2df(collection_name, lat, lon, date_range, bandname, scale, radius = 0.001):
    # unit degree, format:
    # [minlon, minlat,
    #  maxlon, maxlat]
    roi = ee.Geometry.Rectangle([
        lon - radius, lat - radius, 
        lon + radius, lat+ radius
    ])
    start_date, end_date = date_range

    collection = ee.ImageCollection(collection_name)\
        .filterBounds(roi).filterDate(start_date, end_date)
    # Sort the filtered collection by date in ascending order
    collection = collection.sort('system:time_start')

    def interp_image(image, bandname, scale):
        image = ee.Image(image)
        date = image.get('system:time_start')

        image_band = image.select(bandname)
        stats = image_band.reduceRegion(reducer=ee.Reducer.mean(), geometry=roi, scale=scale)
        val = stats.get(bandname)
        return image.set('Info', [date, val])

    # Map the function to the image collection
    # collection = collection.map(interp_image)
    collection = collection.map(lambda image: interp_image(image, bandname, scale))

    # Use aggregate_array to get the values as an array
    array = collection.aggregate_array('Info')


    # Convert the array to a list using getInfo()
    array = array.getInfo()
    df = pd.DataFrame(array, columns = ['DATETIME', 'VALUE'])
    df['DATETIME'] = df['DATETIME'].map(
        lambda x: datetime.utcfromtimestamp(int(x) // 1000)
    )
    df = df.set_index('DATETIME')
    return df

def gee2drive(image, roi, name, folder, scale):
    if not type(roi) == ee.geometry.Geometry:
        minlon = roi[0]
        maxlon = roi[2]
        minlat = roi[1]
        maxlat = roi[3]
        roi = ee.Geometry.BBox(minlon, minlat, maxlon, maxlat)

    task = ee.batch.Export.image.toDrive(
        image=image,
        description=name,
        folder=folder,
        region=roi,
        scale=scale,
        maxPixels = 1e13,
        crs='EPSG:4326'
    )
    task.start()

def gee2local(ee_object, filename, scale, roi, user_params = {}, timeout = 300, proxies = None):
    '''
    Code is from geemap (https://github.com/gee-community/geemap)
    '''
    import os, requests, zipfile, pathlib
    if type(filename) == pathlib.PosixPath:
        filename = filename.as_posix()
    filename_zip = filename.replace(".tif", ".zip")

    params = {
        "dimensions": None,
        "crs": None,
        "crs_transform": None,
        "format": "ZIPPED_GEO_TIFF"

    }
    params["scale"] = scale
    params["region"] = roi
    params.update(user_params)

    try:
        url = ee_object.getDownloadURL(params)
        r = requests.get(url, stream = True, timeout = timeout, proxies = proxies)

        if r.status_code != 200:
            print("An error occurred while downloading.")
            return

        with open(filename_zip, "wb") as fd:
            for chunk in r.iter_content(chunk_size = 1024):
                fd.write(chunk)

    except Exception as e:
        print("An error occurred while downloading.")
        if r is not None:
            print(r.json()["error"]["message"])
        return

    try:
        with zipfile.ZipFile(filename_zip) as z:
            z.extractall(os.path.dirname(filename))
        os.remove(filename_zip)

    except Exception as e:
        print(e)


import os
import ee
import requests
import numpy as np

class EEarth():
    """
    Earth engine object
    """
    def __init__(self, source):
        self.source = source

    def fetch_image(self):
        self.image = ee.Image(self.source)

    def fetch_collection(self, date_range = [], roi = None):
        collection = ee.ImageCollection(self.source)
        # filter by date:
        if date_range:
            collection = collection.filterDate(*date_range)
        # filter by bounds:
        if roi:
            collection = collection.filterBounds(roi)
        self.__length__ = collection.size().getInfo()
        self.collection = collection

class Ecolbox(EEarth):
    """
    Ecolbox: Earth engine Collection Toolbox
    """
    def __init__(self, collection_name, date_range = [], roi = None):
        EEarth.__init__(self, collection_name)
        self.fetch_collection(date_range, roi)
        self.__length__ = self.collection.size().getInfo()

    def __call__(self, collection):
        # update collection box
        self.__length__ = collection.size().getInfo()
        self.collection = collection

    # select certain image
    def __getitem__(self, idx):
        collection = self.collection
        if not isinstance(collection, ee.ee_list.List):
            self.__to_ee_list()
            image = ee.Image(self.__image_list__.get(idx))
        return image

    def __to_ee_list(self):
        self.__image_list__ = self.collection.toList(self.collection.size())

    def to_list(self, collection):
        if not collection:
            collection = self.collection
        return collection.toList(collection.size()).getInfo()

    def get_image_by_index(self, collection = None, idx = 0):
        # default: first.
        if collection:
            collection = collection.toList(collection.size())
            image = ee.Image(collection.get(idx))
        else:
            collection = self.collection
            if not isinstance(collection, ee.ee_list.List):
                self.__to_ee_list()
                image = ee.Image(self.__image_list__.get(idx))
        return image

    def fmap(self, func, *args, auto_update = True, **kwargs):
        # simplify the map function in the future
        if auto_update:
            self.__call__(
                self.collection.map(
                    lambda image: func(image, *args, **kwargs)
                )
            )
        else:
            return self.collection.map(
                lambda image: func(image, *args, **kwargs)
            )

    def reduce_collection(self, collection = None, band = None, label = "DATA", band_reducer = ee.Reducer.mean(), spatial_reducer = ee.Reducer.median()):
        if not collection:
            collection = self.collection
        if band:
            image = collection.select(band).reduce(spatial_reducer).rename(band)
        else:
            image = collection.map(
                    lambda image: image.reduce(band_reducer)
                ).reduce(spatial_reducer).rename(label)
        return image

class Emagebox(object):
    """
    Emagebox: Earth engine Image Toolbox
    """
    def __init__(self, image, scale = 30, max_pixels = 1e8, default_value = -9999):
        self.image = image
        self.scale = scale
        self.max_pixels = max_pixels
        self.default_value = default_value
    
    # select certain band
    def __getitem__(self, band_name):
        return self.image.select(band_name)

    def set_scale(self, scale):
        self.scale = scale
    
    def set_max_pixels(self, max_pixels):
        self.max_pixels = max_pixels
    
    def set_default_value(self, default_value):
        self.default_value = default_value

    def get_band_names(self):
        return self.image.bandNames().getInfo()

    def get_stats(self, roi, reducer = ee.Reducer.mean()):
        stat = self.image.reduceRegion(
            reducer = reducer,
            geometry = roi,
            scale = self.scale,
            maxPixels = self.max_pixels
        )
        return stat.getInfo()
    
    def get_date(self):
        date = ee.Date(self.image.get('system:time_start'))
        return date.format('Y-M-d').getInfo()
    
    def get_proj(self, band, get_info = False):
        # // Get projection information from band.
        proj = self.image.select(band).projection()
        if get_info:
            proj = proj.getInfo()
        return proj

    def get_scale(self, band):
        # // Get scale (in meters) information from band.
        scale = self.image.select(band).projection().nominalScale()
        return scale.getInfo()

    def reproject(self, proj):
        return self.image.reproject(proj)

    def clip(self, roi):
        return self.image.clip(roi)

    def mask_value(self):
        mask = self.image.eq(self.default_value)
        self.image = self.image.updateMask(mask)

    def unmask(self):
        self.image = self.image.unmask(ee.Image.constant(self.default_value))

    def get_value(self, band, point, reducer = ee.Reducer.first()):
        value = self.image.select(band).reduceRegion(reducer, point, self.scale).get(band)
        value = ee.Number(value)
        return value.getInfo()

    def get_values(self, band, roi):
        # getInfo() is limited to 5000 records
        # ee.ee_exception.EEException: Array: No numbers in 'values', must provide a type.
        # see more at: https://gis.stackexchange.com/questions/321560/getting-dem-values-as-numpy-array-in-earth-engine
        # for rectangele: bounds = [-97.94, 26.81, -96.52, 26.84] ## sample land / sea bounds
        # for polygon: bounds = [[[105.532,19.059],[105.606,19.058],[105.605,19.108],[105.530,19.110],[105.532,19.059]]]
        latlng = ee.Image.pixelLonLat().addBands(self.image.clip(roi).unmask(ee.Image.constant(self.default_value)))
        latlng = latlng.reduceRegion(
            reducer = ee.Reducer.toList(), 
            geometry = roi, 
            maxPixels = self.max_pixels, 
            scale = self.scale
        )
        lats = np.array((ee.Array(latlng.get("latitude")).getInfo()))
        lons = np.array((ee.Array(latlng.get("longitude")).getInfo()))
        try:
            values = np.array((ee.Array(latlng.get(band)).getInfo()))
        except:
            values = np.full_like(lats, np.nan, dtype = float)
        # self.values = list(values) ## print as list to check

        if not (values.shape == lats.shape == lons.shape):
            raise Exception(
                f"SizeError: " +
                f"values shape is {values.shape}, " + 
                f"lats shape is {lats.shape}, " + 
                f"lons shape is {lons.shape}."
            )
        return {
            "values":  values,
            "lons": lons,
            "lats": lats
        }

    def localize(self, save_name, save_folder = ".", crs_epsg = "4326"):
        url = self.image.getDownloadURL({
            "name": save_name,
            "crs": "EPSG:" + crs_epsg,
            "scale": self.scale
        })
        if not os.path.exists(save_folder): os.makedirs(save_folder)
        save_dir = f"{save_folder}/{save_name}.zip"

        # Download the subset
        r = requests.get(url, stream = True)
        with open(save_dir, 'wb') as fd:
            for chunk in r.iter_content(chunk_size = 1024):
                fd.write(chunk)


import ee
import folium

class Canvas(object):
	def __init__(self):
		from folium import plugins
		self.cmap = {
			"veg": ['FFFFFF', 'CE7E45', 'DF923D', 'F1B555', 'FCD163', '99B718', '74A901',
					'66A000', '529400', '3E8601', '207401', '056201', '004C00', '023B01',
					'012E01', '011D01', '011301'
					],
			"grey": ['006633', 'E5FFCC', '662A00', 'D8D8D8', 'F5F5F5']
		}
		# Add custom basemaps to folium
		self.basemaps = {
			'Google Maps': folium.TileLayer(
				tiles = 'https://mt1.google.com/vt/lyrs=m&x={x}&y={y}&z={z}',
				attr = 'Google',
				name = 'Google Maps',
				overlay = True,
				control = True
			),
			'Google Satellite': folium.TileLayer(
				tiles = 'https://mt1.google.com/vt/lyrs=s&x={x}&y={y}&z={z}',
				attr = 'Google',
				name = 'Google Satellite',
				overlay = True,
				control = True
			),
			'Google Terrain': folium.TileLayer(
				tiles = 'https://mt1.google.com/vt/lyrs=p&x={x}&y={y}&z={z}',
				attr = 'Google',
				name = 'Google Terrain',
				overlay = True,
				control = True
			),
			'Google Satellite Hybrid': folium.TileLayer(
				tiles = 'https://mt1.google.com/vt/lyrs=y&x={x}&y={y}&z={z}',
				attr = 'Google',
				name = 'Google Satellite',
				overlay = True,
				control = True
			),
			'Esri Satellite': folium.TileLayer(
				tiles = 'https://server.arcgisonline.com/ArcGIS/rest/services/World_Imagery/MapServer/tile/{z}/{y}/{x}',
				attr = 'Esri',
				name = 'Esri Satellite',
				overlay = True,
				control = True
			)
		} 

	def draw(self, images, vis_params, layer_names = [], location=[20, 0], zoom_start = 3, height = 500):
		# draw using folium
		# Add EE drawing method to folium.
		folium.Map.add_ee_layer = self.add_ee_layer

		# Create a folium map object.
		self.m = folium.Map(location = location, zoom_start = zoom_start, height = height)

		# Add custom basemaps
		self.basemaps['Google Maps'].add_to(self.m)
		self.basemaps['Google Satellite Hybrid'].add_to(self.m)
		self.basemaps['Google Terrain'].add_to(self.m)

		# show coors interactively
		self.mouse_position()

		# # click show coors:
		# self._click_popup_coors()

		# click to place a popup marker
		self.m.add_child(folium.ClickForMarker(popup="Hello :)"))

		for image, vis_param, layer_name in zip(images, vis_params, layer_names):
			# Add the elevation model to the map object.
			self.m.add_ee_layer(image, vis_param, layer_name)

		# Add a layer control panel to the map.
		self.m.add_child(folium.LayerControl())

		# Add fullscreen button
		folium.plugins.Fullscreen().add_to(self.m)

		# # Display the map.
		# display(m)

	def save(self, filename):
		# example: m.save("filename.png")
		self.m.save(filename)

	# Define a method for displaying Earth Engine image tiles on a folium map.
	def add_ee_layer(self, ee_object, vis_params, name):
		
		try:    
			# display ee.Image()
			if isinstance(ee_object, ee.image.Image):    
				map_id_dict = ee.Image(ee_object).getMapId(vis_params)
				folium.raster_layers.TileLayer(
				tiles = map_id_dict['tile_fetcher'].url_format,
				attr = 'Google Earth Engine',
				name = name,
				overlay = True,
				control = True
				).add_to(self.m)
			# display ee.ImageCollection()
			elif isinstance(ee_object, ee.imagecollection.ImageCollection):    
				ee_object_new = ee_object.mosaic()
				map_id_dict = ee.Image(ee_object_new).getMapId(vis_params)
				folium.raster_layers.TileLayer(
				tiles = map_id_dict['tile_fetcher'].url_format,
				attr = 'Google Earth Engine',
				name = name,
				overlay = True,
				control = True
				).add_to(self.m)
			# display ee.Geometry()
			elif isinstance(ee_object, ee.geometry.Geometry):    
				folium.GeoJson(
				data = ee_object.getInfo(),
				name = name,
				overlay = True,
				control = True
			).add_to(self.m)
			# display ee.FeatureCollection()
			elif isinstance(ee_object, ee.featurecollection.FeatureCollection):  
				ee_object_new = ee.Image().paint(ee_object, 0, 2)
				map_id_dict = ee.Image(ee_object_new).getMapId(vis_params)
				folium.raster_layers.TileLayer(
				tiles = map_id_dict['tile_fetcher'].url_format,
				attr = 'Google Earth Engine',
				name = name,
				overlay = True,
				control = True
			).add_to(self.m)
		
		except Exception as e:
			print(e)
			print("Could not display {}".format(name))

	def mouse_position(self):
		formatter = "function(num) {return L.Util.formatNum(num, 3) + ' deg';};"

		folium.plugins.MousePosition(
			position='topright',
			separator=' | ',
			empty_string='NaN',
			lng_first=True,
			num_digits=20,
			prefix='Coordinates:',
			lat_formatter=formatter,
			lng_formatter=formatter,
		).add_to(self.m)

	def _click_popup_coors(self):
		folium.LatLngPopup().add_to(self.m)

	# easy visualizaiton parameter generator
	def viz(self, vmin, vmax, palette):
		return {
			"min": vmin,
			"max": vmax,
			"palette": palette
		}


import ee
ee.Initialize()

from . import geeface_lite as gf
__all__ = ["gf"]
from .geeface import *
__all__ += ["EEarth", "Ecolbox", "Emagebox"]
from .utilize import *
__all__ += ["Geometry", "Utils"]
from .canvas import *
__all__ += ["Canvas"]
from .geeface_lite import DataInfo
__all__ += ["DataInfo"]
from .ee_export import ee_export_image, ee_export_image_collection
__all__ += ["ee_export_image", "ee_export_image_collection"]



import ee
import numpy as np
import pandas as pd
from . import chart

def filter_collection(collection: str, date_range: list, bounds: ee.Geometry) -> str:
    """
    Get a list (collection) of ee images filtered by date range and ee geometry (point or so).
    Input:
    -------
    collection: str, ee collection name
    date_range: list, [start: str, end: str]
    bounds: ee.Geometry, e.g.
        ee.Geometry.Polygon(
        [[
        [minlon, maxlat],
        [maxlon, maxlat],
        [maxlon, minlat],
        [minlon, minlat]
        ]]
        ) 
    Return:
    -------
    A list of ee images (collection)
    """
    collection = ee.ImageCollection(collection)
    collection = collection.filterDate(*date_range)
    collection = collection.filterBounds(bounds)
    return collection.toList(collection.size())

def get_collection_size(collection: ee.ImageCollection) -> int:
    """
    Get the length of collection list
    Input:
    -------
    collecton: ee.ImageCollection
    Return:
    -------
    int, size of collection
    """
    return collection.size().getInfo()

def get_image(collection: ee.ImageCollection, count) -> ee.Image:
    """
    Get the ith image from collection
    Input:
    -------
    collection: ee.ImageCollection, a list of images
    count: index of wanted image in collection
    Return:
    -------
    ee.Image
    """
    image = collection.get(count)
    image = ee.Image(image)
    return image

def get_image_bandnames(image: ee.Image) -> list:
    """
    Get the band names of image
    Input:
    -------
    image: ee.Image
    Return:
    -------
    list, a list of band names
    """
    return image.bandNames().getInfo()

def get_proj(image: ee.Image, band: str, pyfmt: bool):
    """
    Get the projection information of an image
    Input:
    -------
    image: ee.Image
    band: str, band name
    pyfmt: bool, python or ee format of output projection
    Return:
    -------
    Projection of image
    """
    proj = image.select(band).projection()
    if pyfmt:
        proj = proj.getInfo()
    return proj
 
def proj_epsg(epsg: int or str = 4326) -> ee.Projection:
    proj = ee.Projection(f'EPSG:{epsg}')
    return proj
 
def get_scale(image: ee.Image) -> float:
    scale = image.projection().nominalScale().getInfo()
    return scale
 
def get_date(image):
    date = ee.Date(image.get('system:time_start'))
    return date.format('Y-M-d HH:mm:ss.SSSS').getInfo()

def timeseries2df(collection: ee.ImageCollection, bands: list, label_bands: list, lon: float, lat: float, date_range: list, scale: float) -> pd.DataFrame:
    """
    Search ee image colection covering point [lon: float, lat:float], and extract time series at the point by date_range ([start: str, end: str])
    Input:
    -------
    collection: ee image collection list
    bands: list, names (str) of bands for ee image
    label_banes: list, user specified names (str) for bands
    lon: float, longitude
    lat: flat, latitude
    date_range: list, [start: str, end: str]
    scale: numeric (float/int), pixel spatial resolution (m)
    Return:
    -------
    df: pandas dataframe time series
    """
    point = ee.Geometry.Point([lon, lat])

    col = ee.ImageCollection(collection).filterBounds(point)
    time_series = col.filterDate(*date_range)
 
    chart_ts = chart.Image.series(**{
        'imageCollection': time_series, 
        'region': point,
        'scale': scale,
        'bands': bands,
        'label_bands':label_bands,
        # 'properties':['CLOUD_COVERAGE_ASSESSMENT'],
        # 'label_properties':['CLOUD_COVER']
    })
 
    df = chart_ts.dataframe
    return df

def region2array(image: ee.Image, roi: ee.Geometry, scale: float, max_pixels: int = 1e9, mask_value: float = -9999) -> pd.DataFrame:
    """
    Get pixel values and coordinates of a region as a dataframe.
    This method converts image with coordinates to dataframe directly, user memory issue may be more likely to occur.
    """
    image = image.addBands(ee.Image.pixelLonLat())
    image = image.unmask(mask_value)
    # ee.Image -> ee.dictionary.Dictionary
    image = image.reduceRegion(
        reducer = ee.Reducer.toList(),
        # crs = 'EPSG:4326',
        geometry = roi,
        maxPixels = 1e13,
        scale = scale
    )
    df = pd.DataFrame.from_dict(image.getInfo(), orient = 'columns')
    return df

def region2arrayB(image: ee.Image, roi: ee.Geometry, band: str, scale: float, max_pixels: int = 1e9, mask_value: float = -9999) -> pd.DataFrame:
    """
    Get pixel values and coordinates of a region as a dataframe.
    Extract pixles values, longitudes, and latitudes seprately, as a plan B for region2array.
    """
    image = image.addBands(ee.Image.pixelLonLat())
    image = image.unmask(mask_value)
    # ee.Image -> ee.dictionary.Dictionary
    image = image.reduceRegion(
        reducer = ee.Reducer.toList(),
        # crs = 'EPSG:4326',
        geometry = roi,
        maxPixels = 1e13,
        scale = scale
    )
    data = np.array((ee.Array(image.get(band)).getInfo()))
    lats = np.array((ee.Array(image.get("latitude")).getInfo()))
    lons = np.array((ee.Array(image.get("longitude")).getInfo()))
    df = pd.DataFrame(zip(data, lats, lons), columns = [band, "latitude", "longitude"])
    return df

def region2arrayC(image: ee.Image, roi: ee.Geometry, bands: list, scale: float, max_pixels: int = 1e9, mask_value: float = -9999) -> pd.DataFrame:
    """
    Get pixel values and coordinates of a region as a dataframe.
    Extract pixles values, longitudes, and latitudes seprately, as a plan B for region2array.
    """
    image = image.addBands(ee.Image.pixelLonLat())
    image = image.unmask(mask_value)
    # ee.Image -> ee.dictionary.Dictionary
    image = image.reduceRegion(
        reducer = ee.Reducer.toList(),
        # crs = 'EPSG:4326',
        geometry = roi,
        maxPixels = 1e13,
        scale = scale
    )
    data_list = []
    for band in bands:
        data = np.array((ee.Array(image.get(band)).getInfo()))
        data_list.append(data)
    lats = np.array((ee.Array(image.get("latitude")).getInfo()))
    lons = np.array((ee.Array(image.get("longitude")).getInfo()))
    df = pd.DataFrame(zip(*data_list, lats, lons), columns = bands + ["latitude", "longitude"])
    return df

def region2array_obs(image: ee.Image, band: str or list, roi: ee.Geometry, proj: ee.Projection, scale: float, max_pixels: int = 1e9, pick_coords: bool = False):
    """
    Obsolete region2array function, will be deprecated.
    """
    if isinstance(band, str):
        band = [band]
    image_array = image.reproject(proj).select(band).reduceRegion(reducer = ee.Reducer.toList(), geometry = roi, scale = scale, maxPixels = max_pixels).getInfo()
    df = pd.DataFrame(list(image_array.values())).T
    df.columns = list(image_array.keys())
    coords = image.pixelLonLat().reproject(proj)
    coords = coords.select(['longitude', 'latitude']).reduceRegion(reducer = ee.Reducer.toList(), geometry = roi, scale = scale, maxPixels = max_pixels).getInfo()
    if pick_coords:
        len_coords = len(coords["longitude"])
        assert len_coords == len(coords["latitude"]), "size of longitude and latitude are different"
        if len_coords > len(df):
            selected_index = np.sort(np.random.choice(np.arange(len_coords), len(df), replace=False))
            coords["longitude"] = np.array(coords["longitude"])[selected_index]
            coords["latitude"] = np.array(coords["latitude"])[selected_index]
        else:
            selected_index = np.sort(np.random.choice(np.arange(len(df)), len_coords, replace=False))
            df = df.iloc[selected_index, :]
    df["lons"] = coords["longitude"]
    df["lats"] = coords["latitude"]
    return df

def draw_image_obs(image: ee.Image, band: str or list, roi: ee.Geometry, to8bit: bool = False, color_order: list = None) -> None:
    """
    Obsolete draw_image function, will be deprecated
    Example:
    import ee
    import numpy as np
    import matplotlib.pyplot as plt
 
 
    # Define an image.
    img = ee.Image('LANDSAT/LC08/C01/T1_SR/LC08_038029_20180810') \
    .select(['B4', 'B5', 'B6'])
 
    # Define an area of interest.
    aoi = ee.Geometry.Polygon(
    [[[-110.8, 44.7],
        [-110.8, 44.6],
        [-110.6, 44.6],
        [-110.6, 44.7]]], None, False)
 
    draw_image(img, ["B4", "B5", "B6"], aoi, to8bit = True, color_order = [2, 1, 0])
    """
    samples = image.sampleRectangle(region=roi)
    if isinstance(band, list):
        array = [np.array(samples.get(b).getInfo())[:, :, np.newaxis] for b in band]
        array = np.concatenate(array, axis = 2) # bands order
    else:
        array = np.array(samples.get(band).getInfo())
    # Scale the data to [0, 255] to show as an RGB image.
    if to8bit:
        array = (255*((array - 100)/3500)).astype('uint8')
        if color_order: 
            array = array[:, :, color_order]
    plt.imshow(array)
    plt.show()

class DataInfo:
    def __init__(self):
      pass
    # --------------------------------------------------------------------------------------------------------------
    @staticmethod
    def s5p_o3(NTRI = True): # start from 2018-09-08
        if NTRI:
            collection = "COPERNICUS/S5P/NRTI/L3_O3"
        else:
            collection = "COPERNICUS/S5P/OFFL/L3_O3"
        bands = [
                "O3_column_number_density", "O3_column_number_density_amf", "O3_slant_column_number_density",
                "O3_effective_temperature", "cloud_fraction", 
                "sensor_azimuth_angle", "sensor_zenith_angle", "solar_azimuth_angle", "solar_zenith_angle"
        ]
        label_bands = [
                    "o3_vcd", "o3_amf", "o3_scd",
                    "o3_temperature", "s5p_cloud_fraction",
                    "s5p_sen_azi", "s5p_sen_zen", "s5p_sun_azi", "s5p_sun_zen"
        ]
        scale = 0.01 * 100 * 1000 # 0.01 arc degree -> meters
        return collection, bands, label_bands, scale
    # --------------------------------------------------------------------------------------------------------------
    @staticmethod
    def s5p_hcho(NTRI = True): # start from 2018-10-02
        if NTRI:
            collection = "COPERNICUS/S5P/NRTI/L3_HCHO"
        else:
            collection = "COPERNICUS/S5P/OFFL/L3_HCHO"
        bands = [
                "tropospheric_HCHO_column_number_density", "tropospheric_HCHO_column_number_density_amf", "HCHO_slant_column_number_density",
                "cloud_fraction",
                "sensor_azimuth_angle", "sensor_zenith_angle", "solar_azimuth_angle", "solar_zenith_angle"
        ]
        label_bands = [
                    "hcho_vcd", "hcho_amf", "hcho_scd",
                    "s5p_cloud_fraction",
                    "s5p_sen_azi", "s5p_sen_zen", "s5p_sun_azi", "s5p_sun_zen"
        ]
        scale = 0.01 * 100 * 1000 # 0.01 arc degree -> meters
        return collection, bands, label_bands, scale
    # --------------------------------------------------------------------------------------------------------------
    @staticmethod
    def s5p_no2(NTRI = True): # start from 2018-07-10
        if NTRI:
            collection = "COPERNICUS/S5P/NRTI/L3_NO2"
        else:
            collection = "COPERNICUS/S5P/OFFL/L3_NO2"
        bands = [
                "NO2_column_number_density", "tropospheric_NO2_column_number_density", "stratospheric_NO2_column_number_density",
                "NO2_slant_column_number_density", "tropopause_pressure", "absorbing_aerosol_index", "cloud_fraction",
                "sensor_altitude", "sensor_azimuth_angle", "sensor_zenith_angle", "solar_azimuth_angle", "solar_zenith_angle"
        ]
        label_bands = [
                    "no2_total_vcd", "no2_vcd", "no2_strato_vcd", "no2_scd",
                    "tropopause_pressure", "absorbing_aerosol_index", "s5p_cloud_fraction",
                    "sp5_sen_alt", "s5p_sen_azi", "s5p_sen_zen", "s5p_sun_azi", "s5p_sun_zen"
        ]
        scale = 0.01 * 100 * 1000 # 0.01 arc degree -> meters
        return collection, bands, label_bands, scale
    # --------------------------------------------------------------------------------------------------------------
    @staticmethod
    def s5p_co(NTRI = True): # start from 2018-06-28
        if NTRI:
            collection = "COPERNICUS/S5P/NRTI/L3_CO"
        else:
            collection = "COPERNICUS/S5P/OFFL/L3_CO"
        bands = [
                "CO_column_number_density", "H2O_column_number_density", "cloud_height",
                "sensor_altitude", "sensor_azimuth_angle", "sensor_zenith_angle", "solar_azimuth_angle", "solar_zenith_angle"
        ]
        label_bands = [
                    "co_vcd", "h2o_vcd", "cloud_height",
                    "sp5_sen_alt", "s5p_sen_azi", "s5p_sen_zen", "s5p_sun_azi", "s5p_sun_zen"
        ]
        scale = 0.01 * 100 * 1000 # 0.01 arc degree -> meters
        return collection, bands, label_bands, scale
    # --------------------------------------------------------------------------------------------------------------
    @staticmethod
    def s5p_so2(NTRI = True): # start from 2018-07-10
        if NTRI:
            collection = "COPERNICUS/S5P/NRTI/L3_SO2"
        else:
            collection = "COPERNICUS/S5P/OFFL/L3_SO2"
        bands = [
                "SO2_column_number_density", "SO2_column_number_density_amf", "SO2_slant_column_number_density",
                "absorbing_aerosol_index", "cloud_fraction",
                "sensor_azimuth_angle", "sensor_zenith_angle", "solar_azimuth_angle", "solar_zenith_angle",
                "SO2_column_number_density_15km"
        ]
        label_bands = [
                    "so2_vcd", "so2_amf", "hcho_scd",
                    "s5p_aai", "s5p_cloud_fraction",
                    "s5p_sen_azi", "s5p_sen_zen", "s5p_sun_azi", "s5p_sun_zen",
                    "so2_vcd_15km"
        ]
        scale = 0.01 * 100 * 1000 # 0.01 arc degree -> meters
        return collection, bands, label_bands, scale
    # --------------------------------------------------------------------------------------------------------------
    @staticmethod
    def s5p_ch4(NTRI = True): # start from 2019-02-08
        if NTRI:
            collection = "COPERNICUS/S5P/NRTI/L3_CH4"
        else:
            collection = "COPERNICUS/S5P/OFFL/L3_CH4"
        bands = [
            "CH4_column_volume_mixing_ratio_dry_air", 
            "aerosol_height", "aerosol_optical_depth",
            "sensor_azimuth_angle", "sensor_zenith_angle",
            "solar_azimuth_angle", "solar_zenith_angle"
        ]
        label_bands = [
                    "xch4", "aerosol_height", "aod", 
                    "s5p_sen_azi", "s5p_sen_zen", 
                    "s5p_sun_azi", "s5p_sun_zen"
        ]
        scale = 0.01 * 100 * 1000 # 0.01 arc degree -> meters
        return collection, bands, label_bands, scale
    # --------------------------------------------------------------------------------------------------------------
    @staticmethod
    def s5p_aai(NTRI = True): # start from 2018-07-04
        if NTRI:
            collection = "COPERNICUS/S5P/NRTI/L3_AER_AI"
        else:
            collection = "COPERNICUS/S5P/OFFL/L3_AER_AI"
        bands = [
            "absorbing_aerosol_index", "sensor_altitude", "sensor_azimuth_angle", "sensor_zenith_angle",
            "solar_azimuth_angle", "solar_zenith_angle"
        ]
        label_bands = [
                    "aai", "sp5_sen_alt", "s5p_sen_azi", "s5p_sen_zen", 
                    "s5p_sun_azi", "s5p_sun_zen"
        ]
        scale = 0.01 * 100 * 1000 # 0.01 arc degree -> meters
        return collection, bands, label_bands, scale
    # --------------------------------------------------------------------------------------------------------------
    @staticmethod
    def modis_reflectance():
        collection = "MODIS/006/MCD43A4" # 1-Day
        bands = [
            'Nadir_Reflectance_Band1', 'Nadir_Reflectance_Band2', 'Nadir_Reflectance_Band3',
            'Nadir_Reflectance_Band4', 'Nadir_Reflectance_Band5', 'Nadir_Reflectance_Band6',
            'Nadir_Reflectance_Band7', 'BRDF_Albedo_Band_Mandatory_Quality_Band1'
        ]
        label_bands = [
            'b1', 'b2', 'b3', 'b4', 'b5', 'b6', 'b7', 'qc'
        ]
        scale = 500
        return collection, bands, label_bands, scale
    # -------------------------------------------------------------------------------------------------------------- 
    @staticmethod
    def modis_aod():
        collection = "MODIS/006/MCD19A2_GRANULES" # 1-Day
        bands = [
                'Optical_Depth_055', 'Optical_Depth_047'
        ]
        label_bands = [
                    'AOD055', 'AOD047'
        ]
        scale = 1000
        return collection, bands, label_bands, scale
    # -------------------------------------------------------------------------------------------------------------- 
    @staticmethod
    def modis_lai_fpar(): # 4-Day
        collection = "MODIS/006/MCD15A3H"
        bands = ['Fpar', 'Lai', "FparLai_QC"]
        label_bands = ['Fpar', 'Lai', "FparLai_QC"]
        scale = 500
        return collection, bands, label_bands, scale
    # -------------------------------------------------------------------------------------------------------------- 
    @staticmethod
    def modis_gpp(sat = "Terra"):
        if sat == "Terra":
            collection = "MODIS/006/MOD17A2H" # 8-Day # Terra; Aqua: MODIS/006/MYD17A2H
        else:
            collection = "MODIS/006/MYD17A2H" # 8-Day Aqua
        bands = [
                'Gpp', 'PsnNet'
        ]
        label_bands = [
                'GPP', 'NEE'
        ]
        scale = 500
        return collection, bands, label_bands, scale
    # -------------------------------------------------------------------------------------------------------------- 
    @staticmethod
    def modis_evi(sat = "Terra"):
        if sat == "Terra":
            collection = "MODIS/MOD09GA_006_EVI" # Daily Terra
        else:
            collection = "MODIS/MYD09GA_006_EVI" # Daily Aqua
        bands = [
                'EVI',
        ]
        label_bands = [
                'EVI'
        ]
        scale = 500
        return collection, bands, label_bands, scale
    # -------------------------------------------------------------------------------------------------------------- 
    @staticmethod
    def modis_ndvi(sat = "Terra"):
        if sat == "Terra":
            collection = "MODIS/MOD09GA_006_NDVI" # Daily Terra
        else:
            collection = "MODIS/MYD09GA_006_NDVI" # Daily Aqua
        bands = [
                'NDVI',
        ]
        label_bands = [
                'NDVI'
        ]
        scale = 500
        return collection, bands, label_bands, scale
    # --------------------------------------------------------------------------------------------------------------
    @staticmethod
    def modis_net_evapo(): # net_evapotranspiration
        collection = "MODIS/006/MOD16A2" # Terra Net Evapotranspiration 8-Day Global 500m
        bands = [
                'ET', 'LE', 'PET', 'PLE', 'ET_QC'
        ]
        label_bands = [
            'ET', 'LE', 'PET', 'PLE', 'ET_QC'
        ]
        scale = 500
        return collection, bands, label_bands, scale
    # --------------------------------------------------------------------------------------------------------------
    @staticmethod
    def landsat7_evi(nday = 8):
        if nday == 8:
            collection = "LANDSAT/LE07/C01/T1_8DAY_EVI" # 8-Day
        elif nday == 32:
            collection = "LANDSAT/LE07/C01/T1_32DAY_EVI" # 32-Day
        else:
            collection = "LANDSAT/LE07/C01/T1_ANNUAL_EVI" # Annual
        bands = [
            'EVI'
        ]
        label_bands = [
            'EVI'
        ]
        scale = 30
        return collection, bands, label_bands, scale

    # --------------------------------------------------------------------------------------------------------------
    @staticmethod
    def landsat7_ndvi(nday = 8):
        if nday == 8:
            collection = "LANDSAT/LE07/C01/T1_8DAY_NDVI" # 8-Day
        elif nday == 32:
            collection = "LANDSAT/LE07/C01/T1_32DAY_NDVI" # 32-Day
        else:
            collection = "LANDSAT/LE07/C01/T1_ANNUAL_NDVI" # Annual
        bands = [
            'NDVI'
        ]
        label_bands = [
            'NDVI'
        ]
        scale = 30
        return collection, bands, label_bands, scale
    # --------------------------------------------------------------------------------------------------------------
    @staticmethod
    def landsat8_evi(nday = 8):
        if nday == 8:
            collection = "LANDSAT/LC08/C01/T1_8DAY_EVI" # 8-Day
        elif nday == 32:
            collection = "LANDSAT/LC08/C01/T1_32DAY_EVI" # 32-Day
        else:
            collection = "LANDSAT/LC08/C01/T1_ANNUAL_EVI" # Annual
        bands = [
            'EVI'
        ]
        label_bands = [
            'EVI'
        ]
        scale = 30
        return collection, bands, label_bands, scale
    # --------------------------------------------------------------------------------------------------------------
    @staticmethod
    def landsat8_ndvi(nday = 8):
        if nday == 8:
            collection = "LANDSAT/LC08/C01/T1_8DAY_NDVI" # 8-Day
        elif nday == 32:
            collection = "LANDSAT/LC08/C01/T1_32DAY_NDVI" # 32-Day
        else:
            collection = "LANDSAT/LC08/C01/T1_ANNUAL_NDVI" # Annual
        bands = [
            'NDVI'
        ]
        label_bands = [
            'NDVI'
        ]
        scale = 30
        return collection, bands, label_bands, scale
    # --------------------------------------------------------------------------------------------------------------
    @staticmethod
    def landsat7_srt1():
        collection = "LANDSAT/LE07/C02/T1_L2"
        bands = ["B1", "B2", "B3", "B4", "B5", "B6", "B7", "sr_atmos_opacity", "sr_cloud_qa"]
        label_bands = ["B1", "B2", "B3", "B4", "B5", "B6", "B7", "sr_atmos_opacity", "sr_cloud_qa"]
        scale = 30
        return collection, bands, label_bands, scale    
    # --------------------------------------------------------------------------------------------------------------
    @staticmethod
    def landsat8_srt1():
        collection = "LANDSAT/LC08/C02/T1_L2"
        bands = ['B1', 'B2', 'B3', 'B4', 'B5', 'B6', 'B7', 'B10', 'B11', 'sr_aerosol']
        label_bands = ['B1', 'B2', 'B3', 'B4', 'B5', 'B6', 'B7', 'B10', 'B11', 'sr_aerosol']
        scale = 30
        return collection, bands, label_bands, scale
    # --------------------------------------------------------------------------------------------------------------
    @staticmethod
    def get_cfsv2():
        collection = 'NOAA/CFSV2/FOR6H' # 6-Hourly
        bands = [
                'Temperature_height_above_ground', 'Downward_Short-Wave_Radiation_Flux_surface_6_Hour_Average', 'Volumetric_Soil_Moisture_Content_depth_below_surface_layer_5_cm',
                'Maximum_specific_humidity_at_2m_height_above_ground_6_Hour_Interval', 'Minimum_specific_humidity_at_2m_height_above_ground_6_Hour_Interval',
                'u-component_of_wind_height_above_ground', 'v-component_of_wind_height_above_ground', 'Precipitation_rate_surface_6_Hour_Average'
        ]
        label_bands = [
                "TAIR", "RG", "SMP", 
                "SHMAX", "SHMIN", "UWIND", "VWIND", "PRECIP"
        ]
        scale = 0.2*100*1000
        return collection, bands, label_bands, scale
    # --------------------------------------------------------------------------------------------------------------
    @staticmethod
    def get_cloud():
        collection = "NOAA/CDR/PATMOSX/V53"
        bands = ["cloud_fraction", "cloud_fraction_uncertainty", "cloud_probability"]
        label_bands = ["cloud_fraction", "cloud_fraction_uncertainty", "cloud_probability"]
        scale = 0.1 * 100 * 1000
        return collection, bands, label_bands, scale
    # --------------------------------------------------------------------------------------------------------------
    @staticmethod
    def get_toms_omi_ozone():
        collection = "TOMS/MERGED"
        bands = ["ozone"]
        label_bands = ["ozone"]
        scale = 1 * 100 * 1000
        return collection, bands, label_bands, scale
    # --------------------------------------------------------------------------------------------------------------
    @staticmethod
    def pml_v2_water():
        collection = "CAS/IGSNRR/PML/V2"
        bands = ["GPP", "Ec", "Es", "Ei", "ET_water"]
        label_bands = ["GPP", "transpiration", "evaporation", "Interception_from_vegetation_canopy", "ET_water"]
        scale = 500
        return collection, bands, label_bands, scale

    # --------------------------------------------------------------------------------------------------------------
    @staticmethod
    def VIIRS_nightlight_corrected():
        collection = "NOAA/VIIRS/DNB/MONTHLY_V1/VCMSLCFG" # 1-Day
        bands = [
            "avg_rad", # nanoWatts/cm2/sr; Average DNB radiance values.
            "cf_cvg" # Cloud-free coverages; the total number of observations that went into each pixel. 
                    # This band can be used to identify areas with low numbers of observations where the quality is reduced.
        ]
        label_bands = [
            "avg_rad",
            "cf_cvg"
        ]
        scale = 450 # 15 arc seconds 1 arc second = 30 meters
        return collection, bands, label_bands, scale

    # --------------------------------------------------------------------------------------------------------------
    @staticmethod
    # Sentinel-2 MSI Level-1C
    def s2_msi1C():
        collection = "COPERNICUS/S2"
        bands = [
            'B1', 'B2', 'B3', 'B4', 'B5', 'B6', 'B7', 
            'B8', 'B8A', 'B9', 'B10', 'B11', 'B12',
            'QA10', 'QA20', 'QA60'
        ]
        label_bands = [
            'B1', 'B2', 'B3', 'B4', 'B5', 'B6', 'B7', 
            'B8', 'B8A', 'B9', 'B10', 'B11', 'B12',
            'QA10', 'QA20', 'QA60'
        ]

        scale = 10 # 10 m for most bands
        return collection, bands, label_bands, scale

    # --------------------------------------------------------------------------------------------------------------
    @staticmethod
    # Sentinel-2 MSI Level-2A
    def s2_msi2A():
        collection = "COPERNICUS/S2_SR"
        bands = [
            'B1', 'B2', 'B3', 'B4', 'B5', 'B6', 'B7', 
            'B8', 'B8A', 'B9', 'B11', 'B12',
            'AOT', 'WVP', 'SCL', 'TCI_R', 'TCI_G', 'TCI_B',
            'MSK_CLDPRB', 'MSK_SNWPRB', 
            'QA10', 'QA20', 'QA60'
        ]
        label_bands = [
            'B1', 'B2', 'B3', 'B4', 'B5', 'B6', 'B7', 
            'B8', 'B8A', 'B9', 'B11', 'B12',
            'AOT', 'WVP', 'SCL', 'TCI_R', 'TCI_G', 'TCI_B',
            'MSK_CLDPRB', 'MSK_SNWPRB', 
            'QA10', 'QA20', 'QA60'
        ]

        scale = 10 # 10 m for most bands
        return collection, bands, label_bands, scale
    
    # --------------------------------------------------------------------------------------------------------------
    @staticmethod
    # Sentinel-3 OLCI dataset
    def s3_olci():
        collection = "COPERNICUS/S3/OLCI"
        bands = [
                'Oa01_radiance', 'Oa02_radiance', 'Oa03_radiance', 'Oa04_radiance', 'Oa05_radiance',
                'Oa06_radiance', 'Oa07_radiance', 'Oa08_radiance', 'Oa09_radiance', 'Oa10_radiance',
                'Oa11_radiance', 'Oa12_radiance', 'Oa13_radiance', 'Oa14_radiance', 'Oa15_radiance',
                'Oa16_radiance', 'Oa17_radiance', 'Oa18_radiance', 'Oa19_radiance', 'Oa20_radiance', 'Oa21_radiance'
        ]
        label_bands = ['oa01', 'oa02', 'oa03', 'oa04', 'oa05',
                    'oa06', 'oa07', 'oa08', 'oa09', 'oa10',
                    'oa11', 'oa12', 'oa13', 'oa14', 'oa15',
                    'oa16', 'oa17', 'oa18', 'oa19', 'oa20', 'oa21'
        ]

        # rad_scale * band -> radiace
        # in order: band1 -> band21
        rad_scale = [
                    0.0139465, 0.0133873, 0.0121481, 0.0115198, 0.0100953,
                    0.0123538, 0.00879161, 0.00876539, 0.0095103, 0.00773378,
                    0.00675523, 0.0071996, 0.00749684, 0.0086512, 0.00526779,
                    0.00530267, 0.00493004, 0.00549962, 0.00502847, 0.00326378, 0.00324118
        ]
        scale = 300
        return collection, bands, label_bands, scale, rad_scale

    # --------------------------------------------------------------------------------------------------------------
    @staticmethod
    def AVHRR_reflectance():
        collection = "NOAA/CDR/AVHRR/SR/V5"
        bands = [
            'SREFL_CH1', 'SREFL_CH2', 'SREFL_CH3', 'BT_CH3', 'BT_CH4', 'BT_CH5'
        ]
        label_bands = [
            'SREFL_CH1', 'SREFL_CH2', 'SREFL_CH3', 'BT_CH3', 'BT_CH4', 'BT_CH5'
        ]
        scale = 5566
        return collection, bands, label_bands, scale

    # --------------------------------------------------------------------------------------------------------------
    @staticmethod
    def ERA5_hourly():
        # ERA5-Land Hourly - ECMWF Climate Reanalysis
        collection = "ECMWF/ERA5_LAND/HOURLY"
        bands  = [
            'surface_solar_radiation_downwards', 'surface_thermal_radiation_downwards',
            'dewpoint_temperature_2m', 'temperature_2m', 'soil_temperature_level_1', 
            'u_component_of_wind_10m', 'v_component_of_wind_10m', 
            'total_precipitation', 'surface_pressure',
        ]
        label_bands = [
            'SWIN', 'LWIN', 'D2m', 'T2m', 'Tsoil1', 'U', 'V', 'P', 'PA'
        ]
        scale = 11132
        return collection, bands, label_bands, scale

    # --------------------------------------------------------------------------------------------------------------
    @staticmethod
    def modis_lc(): # 2001-01-01T00:00:00Z - 2019-01-01T00:00:00
        collection = "MODIS/006/MCD12Q1"
        bands = [
            "LC_Type1", "LC_Type2", "LC_Type3", "LC_Type4", "LC_Type5", 
            "LC_Prop1_Assessment", "LC_Prop2_Assessment", "LC_Prop3_Assessment", 
            "LC_Prop1", "LC_Prop2", "LC_Prop3", 
            "QC", "LW"
        ]
        label_bands = [
            "LC_Type1", "LC_Type2", "LC_Type3", "LC_Type4", "LC_Type5", 
            "LC_Prop1_Assessment", "LC_Prop2_Assessment", "LC_Prop3_Assessment", 
            "LC_Prop1", "LC_Prop2", "LC_Prop3", 
            "QC", "LW"
        ]
        scale = 500 # meters
        return collection, bands, label_bands, scale


import ee
import pandas as pd
from datetime import datetime

def gee2df(collection_name, lat, lon, date_range, bandname, scale, radius = 0.001):
    # unit degree, format:
    # [minlon, minlat,
    #  maxlon, maxlat]
    roi = ee.Geometry.Rectangle([
        lon - radius, lat - radius, 
        lon + radius, lat+ radius
    ])
    start_date, end_date = date_range

    collection = ee.ImageCollection(collection_name)\
        .filterBounds(roi).filterDate(start_date, end_date)
    # Sort the filtered collection by date in ascending order
    collection = collection.sort('system:time_start')

    def interp_image(image, bandname, scale):
        image = ee.Image(image)
        date = image.get('system:time_start')

        image_band = image.select(bandname)
        stats = image_band.reduceRegion(reducer=ee.Reducer.mean(), geometry=roi, scale=scale)
        val = stats.get(bandname)
        return image.set('Info', [date, val])

    # Map the function to the image collection
    # collection = collection.map(interp_image)
    collection = collection.map(lambda image: interp_image(image, bandname, scale))

    # Use aggregate_array to get the values as an array
    array = collection.aggregate_array('Info')


    # Convert the array to a list using getInfo()
    array = array.getInfo()
    df = pd.DataFrame(array, columns = ['DATETIME', 'VALUE'])
    df['DATETIME'] = df['DATETIME'].map(
        lambda x: datetime.utcfromtimestamp(int(x) // 1000)
    )
    df = df.set_index('DATETIME')
    return df


#!/usr/bin/env python
#-*- coding:utf-8 -*-

#############################################
# File Name: setup.py
# Author: Songyan Zhu
# Mail: soonyenju@foxmail.com
# Created Time:  2018-10-23 13:28:34
#############################################


from setuptools import setup, find_packages

setup(
	name = "scigee",
	version = "0.0.15",
	keywords = ("easy geo warpper", "atmospheric data","satellite data", "flux"),
	description = "For faster proccessing geofile",
	long_description = "Read/write and process rs/gis related data, especially atmospheric rs data.",
	license = "MIT Licence",

	url="https://github.com/soonyenju/scigee",
	author = "Songyan Zhu",
	author_email = "soonyenju@foxmail.com",

	packages = find_packages(),
	include_package_data = True,
	platforms = "any",
	install_requires=[
            # "geopandas==0.4.1",
            # "scipy==1.2.1",
            # "GDAL==2.3.3",
            # "pyproj==1.9.6",
            # "Shapely==1.6.4.post1",
            # "rasterio==1.0.21",
            # "numpy==1.16.0",
			# "geopandas",
			# "rasterio"
			"pygal",
			"geetools"

	]
)


# coding=utf-8

""" 
https://github.com/giswqs/geemap/blob/master/AUTHORS.rst
Package: geemap
Author: Qiusheng Wu, qwu18@utk.edu
"""

import ee
import os
import requests
import shutil
import tarfile
import urllib.request
import zipfile

def ee_export_image(
    ee_object, filename, scale=None, crs=None, region=None, file_per_band=False
):
    """Exports an ee.Image as a GeoTIFF.
    Args:
        ee_object (object): The ee.Image to download.
        filename (str): Output filename for the exported image.
        scale (float, optional): A default scale to use for any bands that do not specify one; ignored if crs and crs_transform is specified. Defaults to None.
        crs (str, optional): A default CRS string to use for any bands that do not explicitly specify one. Defaults to None.
        region (object, optional): A polygon specifying a region to download; ignored if crs and crs_transform is specified. Defaults to None.
        file_per_band (bool, optional): Whether to produce a different GeoTIFF per band. Defaults to False.
    """

    if not isinstance(ee_object, ee.Image):
        print("The ee_object must be an ee.Image.")
        return

    filename = os.path.abspath(filename)
    basename = os.path.basename(filename)
    name = os.path.splitext(basename)[0]
    filetype = os.path.splitext(basename)[1][1:].lower()
    filename_zip = filename.replace(".tif", ".zip")

    if filetype != "tif":
        print("The filename must end with .tif")
        return

    try:
        print("Generating URL ...")
        params = {"name": name, "filePerBand": file_per_band}
        if scale is None:
            scale = ee_object.projection().nominalScale().multiply(10)
        params["scale"] = scale
        if region is None:
            region = ee_object.geometry()
        params["region"] = region
        if crs is not None:
            params["crs"] = crs

        try:
            url = ee_object.getDownloadURL(params)
        except Exception as e:
            print("An error occurred while downloading.")
            print(e)
            return
        print(f"Downloading data from {url}\nPlease wait ...")
        r = requests.get(url, stream=True)

        if r.status_code != 200:
            print("An error occurred while downloading.")
            return

        with open(filename_zip, "wb") as fd:
            for chunk in r.iter_content(chunk_size=1024):
                fd.write(chunk)

    except Exception as e:
        print("An error occurred while downloading.")
        print(r.json()["error"]["message"])
        return

    try:
        with zipfile.ZipFile(filename_zip) as z:
            z.extractall(os.path.dirname(filename))
        os.remove(filename_zip)

        if file_per_band:
            print(f"Data downloaded to {os.path.dirname(filename)}")
        else:
            print(f"Data downloaded to {filename}")
    except Exception as e:
        print(e)

def ee_export_image_collection(
    ee_object, out_dir, scale=None, crs=None, region=None, file_per_band=False
):
    """Exports an ImageCollection as GeoTIFFs.
    Args:
        ee_object (object): The ee.Image to download.
        out_dir (str): The output directory for the exported images.
        scale (float, optional): A default scale to use for any bands that do not specify one; ignored if crs and crs_transform is specified. Defaults to None.
        crs (str, optional): A default CRS string to use for any bands that do not explicitly specify one. Defaults to None.
        region (object, optional): A polygon specifying a region to download; ignored if crs and crs_transform is specified. Defaults to None.
        file_per_band (bool, optional): Whether to produce a different GeoTIFF per band. Defaults to False.
    """

    if not isinstance(ee_object, ee.ImageCollection):
        print("The ee_object must be an ee.ImageCollection.")
        return

    if not os.path.exists(out_dir):
        os.makedirs(out_dir)

    try:

        count = int(ee_object.size().getInfo())
        print(f"Total number of images: {count}\n")

        for i in range(0, count):
            image = ee.Image(ee_object.toList(count).get(i))
            name = image.get("system:index").getInfo() + ".tif"
            filename = os.path.join(os.path.abspath(out_dir), name)
            print(f"Exporting {i + 1}/{count}: {name}")
            ee_export_image(
                image,
                filename=filename,
                scale=scale,
                crs=crs,
                region=region,
                file_per_band=file_per_band,
            )
            print("\n")

    except Exception as e:
        print(e)


import ee

class VI(object):
    def __init__(self):
        pass

    def __call__(self, image, vi, **kwargs):
        if vi == "ndvi":
            image = self.calc_ndvi(image, **kwargs)
        elif vi == "evi":
            image = self.calc_evi(image, **kwargs)
        elif vi == "cire":
            image = self.calc_cire(image, **kwargs)
        elif vi == "ndi":
            image = self.calc_ndi(image, **kwargs)

        return image

    @staticmethod
    def calc_ndvi(image, **kwargs):
        nir = kwargs["nir"]
        red = kwargs["red"]

        evi = image.expression(
            '((NIR - RED) / (NIR + RED))', {
            'NIR': image.select(nir),
            'RED': image.select(red)
        })
        return image.addBands(evi.rename("NDVI"))

    @staticmethod
    def calc_evi(image, **kwargs):
        nir = kwargs["nir"]
        red = kwargs["red"]
        blue = kwargs["blue"]

        evi = image.expression(
            '2.5 * ((NIR - RED) / (NIR + 6 * RED - 7.5 * BLUE + 1))', {
            'NIR': image.select(nir),
            'RED': image.select(red),
            'BLUE': image.select(blue)
        })
        return image.addBands(evi.rename("EVI"))

    @staticmethod
    def calc_cire(image, **kwargs):
        re2 = kwargs["re2"]
        re3 = kwargs["re3"]

        cire = image.expression(
            '(RE3/RE2) - 1', {
            'RE2': image.select(re2),
            'RE3': image.select(re3)
        })
        return image.addBands(cire.rename("CIRE"))

    @staticmethod
    def calc_ndi(image, **kwargs):
        band_a = kwargs["b1"]
        band_b = kwargs["b2"]

        ndi = image.expression(
            '(band_a - band_b) / (band_a + band_b)', {
            'band_a': image.select(band_a),
            'band_b': image.select(band_b)
        })
        if "label" in kwargs.keys():
            label = kwargs["label"]
        else:
            label = "NDI"
        return image.addBands(ndi.rename(label))

class Point(object):
    def __init__(self):
        pass

    def __call__(self, bounds = None, lon = None, lat = None):
        if bounds:
            lon, lat = bounds
        assert lon, "No longitude input..."
        assert lat, "No latitude input..."
        point = ee.Geometry.Point(lon, lat)
        return point

    @classmethod
    def get_circle_buffer(self, point, buffer_size = 100):
        return point.buffer(buffer_size)

    @classmethod
    def get_rect_buffer(self, point, buffer_size = 0.5):
        # buffer_size unit eq proj unit, e.g., degree for WGS84
        lon, lat = point.getInfo()["coordinates"]
        # example: [-97.94, 26.81, -96.52, 26.84]
        bounds = [lon - buffer_size, lat - buffer_size, lon + buffer_size, lat + buffer_size]
        return ee.Geometry.Rectangle(bounds)

class Polygon(object):
    def __init__(self):
        pass

    def __call__(self, bounds):
        # for rectangele: bounds = [-97.94, 26.81, -96.52, 26.84] ## sample land / sea bounds
        # for polygon: bounds = [[[105.532,19.059],[105.606,19.058],[105.605,19.108],[105.530,19.110],[105.532,19.059]]]

        if isinstance(bounds, list):
            if np.array(bounds).ndim == 1:
                region = ee.Geometry.Rectangle(bounds)
            else:
                region = ee.Geometry.Polygon(bounds)
        return region

class Geometry(Point, Polygon):
    def __init__(self, bounds = None, lon = None, lat = None):
        Point.__init__(self)
        Polygon.__init__(self)
        self.bounds = bounds
        self.lon = lon
        self.lat = lat

    def __call__(self, geom_type):
        if geom_type == 0 or geom_type == "point":
            geom = Point.__call__(self, bounds = self.bounds, lon = self.lon, lat = self.lat)
        elif geom_type == 2 or geom_type == "polygon":
            geom = Polygon.__call__(self, bounds = self.bounds)
        return geom


class Utils(VI):
    def __init__(self):
        VI.__init__(self)

    @classmethod
    def calc_vi(self, image, vi, **kwargs):
        return VI.__call__(self, image, vi, **kwargs)

    @classmethod
    def sentinel_2_cloud_mask(self, image, mask_out = True):
        """
        javascript code:

        # function maskS2clouds(image) {
        #   var qa = image.select('QA60');

        #   // Bits 10 and 11 are clouds and cirrus, respectively.
        #   var cloudBitMask = 1 << 10;
        #   var cirrusBitMask = 1 << 11;

        #   // Both flags should be set to zero, indicating clear conditions.
        #   var mask = qa.bitwiseAnd(cloudBitMask).eq(0)
        #       .and(qa.bitwiseAnd(cirrusBitMask).eq(0));

        #   return image.updateMask(mask).divide(10000);
        # }

        European Space Agency (ESA) clouds from 'QA60', i.e. Quality Assessment band at 60m
        
        parsed by Nick Clinton
        """

        qa = image.select('QA60')

        # bits 10 and 11 are clouds and cirrus
        cloudBitMask = int(2**10)
        cirrusBitMask = int(2**11)

        # both flags set to zero indicates clear conditions.
        mask = qa.bitwiseAnd(cloudBitMask).eq(0).And(\
            qa.bitwiseAnd(cirrusBitMask).eq(0))
        
        if mask_out:
            return image.updateMask(mask)#.divide(10000)
        else:
            # clouds is not clear
            cloud = mask.Not().rename(['ESA_clouds'])

            # return the masked and scaled data.
            return image.addBands(cloud)

    # @classmethod
    # def coors2roi(self, bounds):
    #     # for rectangele: bounds = [-97.94, 26.81, -96.52, 26.84] ## sample land / sea bounds
    #     # for polygon: bounds = [[[105.532,19.059],[105.606,19.058],[105.605,19.108],[105.530,19.110],[105.532,19.059]]]
    #     if isinstance(bounds, list):
    #         if np.array(bounds).ndim == 1:
    #             roi = ee.Geometry.Rectangle(bounds)
    #         else:
    #             roi = ee.Geometry.Polygon(bounds)
    #     return roi

    # @classmethod
    # def coor2point(self, coors = None, lon = None, lat = None):
    #     if coors:
    #         lon, lat = coors
    #     assert lon, "No longitude input..."
    #     assert lat, "No latitude input..."
    #     point = ee.Geometry.Point(lon, lat)
    #     return point

    # @classmethod
    # def get_circle_buffer(self, point, buffer_size = 100):
    #     return point.buffer(buffer_size)

    # @classmethod	
    # def get_rect_buffer(self, point, buffer_size = 0.5):
    #     # buffer_size unit eq proj unit, e.g., degree for WGS84
    #     lon, lat = point.getInfo()["coordinates"]
    #     # example: [-97.94, 26.81, -96.52, 26.84]
    #     bounds = [lon - buffer_size, lat - buffer_size, lon + buffer_size, lat + buffer_size]
    #     return ee.Geometry.Rectangle(bounds)